package com.example.fuzzControll.tools.test;

import com.alibaba.fastjson.JSONObject;
import com.example.fuzzControll.annotion.NeedCutBefore;
import com.example.fuzzControll.domain.bo.CmdStartParams;
import com.example.fuzzControll.domain.po.MissionInfo;
import com.example.fuzzControll.domain.vo.KittyResult;
import com.example.fuzzControll.domain.vo.TestReturnEntity;
import com.example.fuzzControll.exception.testException.AflnetException;
import com.example.fuzzControll.exception.testException.CmdException;
import com.example.fuzzControll.exception.mysqlException.MysqlException;
import com.example.fuzzControll.service.impl.websocketClientServiceImpl;
import com.example.fuzzControll.tools.system.GlobalClass;
import com.example.fuzzControll.tools.system.SystemRunningParams;
import lombok.extern.slf4j.Slf4j;
import org.springframework.stereotype.Component;
import org.springframework.transaction.annotation.Transactional;

import java.io.*;

import java.util.*;
import java.util.concurrent.ConcurrentHashMap;
import java.util.stream.Collectors;


//todo need modify
@Slf4j
@Component("TestCmdTools")
public class TestCmdTools {
    Boolean send = false;

    Boolean show = true;

    /**
     * 运行指令返回错误信息
     */
    public List<String> runCmdReturnError(String cmd, String caller) throws CmdException {
        List<String> result = new ArrayList<String>();
        List<String> error = new ArrayList<String>();
        try {
            log.info(caller + " is running!");
            log.info("Running cmd:[{}]", cmd);
            Process process = Runtime.getRuntime().exec(cmd);
            printMessage(process.getInputStream(), result);
            printMessage(process.getErrorStream(), error);
            process.waitFor();
            log.info(caller + " end!");
        } catch (Exception e) {
            e.printStackTrace();
            throw new CmdException(caller + " run cmd failed!");
        }
        return error;
    }

    /**
     * 运行会阻塞的命令行
     */
    public List<String> runClogCmd(String cmd, List<String> response, String caller) throws CmdException {
        try {
            log.info(caller + " is running!");
            log.info("Running cmd:[{}]", cmd);
            Process process = Runtime.getRuntime().exec(cmd);
            printMessageClogCmd(process.getInputStream(), response);
            printMessageClogCmd(process.getErrorStream(), new ArrayList<>());
            process.waitFor();
            log.info(caller + " end!");
        } catch (Exception e) {
            e.printStackTrace();
            throw new CmdException(caller + " run cmd failed!");
        }
        return response;
    }

    private void printMessageClogCmd(InputStream inputStream, List<String> response) {
        new Thread(new Runnable() {
            @Override
            public void run() {
                Reader reader = new InputStreamReader(inputStream);
                BufferedReader bf = new BufferedReader(reader);
                String line = null;
                try {
                    while ((line = bf.readLine()) != null) {
                        log.info(line);
                        response.add(line);
                    }
                } catch (IOException e) {
                    e.printStackTrace();
                }

            }
        }).start();
    }

    /**
     * 运行不需要后台运行cmd
     */
    public List<String> runCmd(String cmd, String caller) throws CmdException {
        List<String> result = new ArrayList<String>();
        List<String> error = new ArrayList<String>();
        try {
            log.info(caller + " is running!");
            log.info("Running cmd:[{}]", cmd);
            Process process = Runtime.getRuntime().exec(cmd);
            printMessage(process.getInputStream(), result);
            printMessage(process.getErrorStream(), error);
            process.waitFor();
            log.info(caller + " end!");
        } catch (Exception e) {
            e.printStackTrace();
            throw new CmdException(caller + " run cmd failed!");
        }
        return result;
    }
    //todo 不同协议种子路径也不同

    /**
     * 运行需要后台运行cmd
     * 通过websocket返回数据
     */
    public void runProgramCmd(String cmd, String outputFileName, MissionInfo missionInfo) throws AflnetException {
        try {
            Process process = Runtime.getRuntime().exec(cmd);//执行模糊测试指令
            /*打印模糊测试输出*/
            printMessageToWeb(process.getInputStream());
            printMessage(process.getErrorStream(), new ArrayList<String>());
            process.waitFor();
            log.info("Aflnet cmd have been run.");
        } catch (Exception e) {
            log.error("alfnet run error!:" + e.getMessage());
            throw new AflnetException("");
        }
    }

    /**
     * 运行需要后台运行cmd
     * 将数据存入文件中
     */
    public Map<String, List<String>> runProgramCmdAndResult(String cmd, String caller, String missionName) throws
            CmdException {
        Map<String, List<String>> result = new HashMap();
        List<String> out = Collections.synchronizedList(new ArrayList<String>());
        List<String> error = Collections.synchronizedList(new ArrayList<String>());
        try {
            Process process = Runtime.getRuntime().exec(cmd);
            printMessageByProgramCmd(process.getInputStream(), out);
            printMessageByProgramCmd(process.getErrorStream(), error);
            process.waitFor();
        } catch (Exception e) {
            e.printStackTrace();
            throw new CmdException("run fuzz error !");
        }
        out = out.stream().map(s -> s + System.lineSeparator()).collect(Collectors.toList());
        error = error.stream().map(s -> s + System.lineSeparator()).collect(Collectors.toList());
        result.put("out", out);
        result.put("error", error);
        //todo 对于无法运行和数据较长的还需要观察
        /*新开一个线程存入数据*/
        List<String> finalOut = out;
        List<String> finalError = error;
        new Thread(new Runnable() {
            @Override
            public void run() {
                dataBackUpTransaction(caller, finalOut, finalError, missionName);
            }
        }).start();
        //todo 需要传入任务参数，后期再开发
        return result;
    }

    private List<String> printMessageByProgramCmd(final InputStream input, final List<String> result) {
        new Thread(new Runnable() {
            @Override
            public void run() {
                Reader reader = new InputStreamReader(input);
                BufferedReader bf = new BufferedReader(reader);
                String line = null;
                try {
                    while ((line = bf.readLine()) != null) {
                        result.add(line);
                    }
                } catch (IOException e) {
                    e.printStackTrace();
                }

            }
        }).start();
        return result;
    }

    /**
     * 错误输出
     *
     * @param input
     * @param result
     * @return
     */
    private List<String> printMessage(final InputStream input, final List<String> result) {
        new Thread(new Runnable() {
            @Override
            public void run() {
                Reader reader = new InputStreamReader(input);
                BufferedReader bf = new BufferedReader(reader);
                String line = null;
                try {
                    while ((line = bf.readLine()) != null) {
                        result.add(line);
                    }
                } catch (IOException e) {
                    e.printStackTrace();
                }

            }
        }).start();
        return result;
    }

    private void printMessageToWeb(final InputStream input) throws IOException, AflnetException {
        show = true;
        Reader reader = new InputStreamReader(input);
        BufferedReader bf = new BufferedReader(reader);
        String line = null;
        TestReturnEntity returnEntity = new TestReturnEntity();
        while ((line = bf.readLine()) != null && TestControlTools.getIsRunning()) {
            makeReturnEntity(line, returnEntity);
            if (send) {
                String data = JSONObject.toJSONString(returnEntity);
                try {
                    new Thread(new Runnable() {
                        @Override
                        public void run() {
                            dataToMap(data);
                        }
                    }).start();
                    websocketClientServiceImpl.webSocketClient.send(data);
                } catch (Exception e) {
                    log.error("Web is disconnected:[{}]", e.getMessage());
                }
            }
        }
    }

    private void dataToMap(String data) {
        try {
            JSONObject dataJson = JSONObject.parseObject(data);
            LinkedHashMap<String, String> testProgressMap = SystemRunningParams.TestProgressMap;
            testProgressMap.put("系统已运行时间", dataJson.getString("run_time"));
            testProgressMap.put("距上次发现新路径过去多少时间", dataJson.getString("last_new_path"));
            testProgressMap.put("据上次发现唯一崩溃过去多少时间", dataJson.getString("last_uniq_crash"));
            testProgressMap.put("距上次发现唯一挂起过去多少时间", dataJson.getString("last_uniq_hang"));
            testProgressMap.put("队列循环次数", dataJson.getString("cycles_done"));
            testProgressMap.put("队列测试用例总数", dataJson.getString("total_paths"));
            testProgressMap.put("添加到队列中的用例数量", dataJson.getString("own_finds"));
            testProgressMap.put("队列中待fuzz的用例数量", dataJson.getString("pending"));
            testProgressMap.put("发现的唯一崩溃的总数", dataJson.getString("uniq_crashes"));
            testProgressMap.put("发现的唯一挂起总数", dataJson.getString("uniq_hangs"));
            testProgressMap.put("当前正在fuzz队列的第几个测试用例/测试进度", dataJson.getString("now_processing"));
            testProgressMap.put("放弃的用例数量/占总用例的比率", dataJson.getString("paths_timed_out"));
            testProgressMap.put("当前正在执行的变异阶段", dataJson.getString("now_trying"));
            testProgressMap.put("当前执行次数/该阶段需要执行的次数（执行进度）", dataJson.getString("stage_execs"));
            testProgressMap.put("总执行次数", dataJson.getString("total_crashes"));
            testProgressMap.put("平均执行速度", dataJson.getString("exec_speed"));
            testProgressMap.put("总的崩溃数量", dataJson.getString("total_crashes"));
            testProgressMap.put("总的超时数量", dataJson.getString("total_tmouts"));
        } catch (Exception e) {
            log.error("There is no key:{}", e.getMessage());
            throw new RuntimeException();
        }
    }

    /**
     * aflnet模糊测试数据键值对返回
     *
     * @param line
     * @param returnEntity
     * @return
     */
    private TestReturnEntity makeReturnEntity(String line, TestReturnEntity returnEntity) {
//        if (line.contains("All set and ready to roll!") || line.contains("american fuzzy") || line.contains("process timing  overall results")) {
//            show = false;
//        } else if (line.contains("PROGRAM ABORT")) {
//            log.error("Aflnet run failed !");
//            throw new AflnetException("");
//        }
        if (show) {
            log.info(line);
        }
        SystemRunningParams.testTimeMessage.get("aflnet").put("start", System.currentTimeMillis());
        try {
            if (line.contains("run time")) {
                send = false;
                int run_time = line.indexOf(":");
                int cycles_done = line.indexOf("cycles done :");
                returnEntity.setRun_time(line.substring(run_time + 1, run_time + 30));
                returnEntity.setCycles_done(line.substring(cycles_done + 13, cycles_done + 18));
            } else if (line.contains("last new path")) {
                int last_new_path = line.indexOf(":");
                int total_paths = line.indexOf("total paths :");
                returnEntity.setLast_new_path(line.substring(last_new_path + 1, last_new_path + 30));
                returnEntity.setTotal_paths(line.substring(total_paths + 13, line.indexOf("total paths :") + 16));
            } else if (line.contains("last uniq crash")) {
                int last_uniq_crash = line.indexOf(":");
                int uniq_crashes = line.indexOf("uniq crashes :");
                returnEntity.setLast_uniq_crash(line.substring(last_uniq_crash + 1, last_uniq_crash + 30));
                returnEntity.setUniq_crashes(line.substring(uniq_crashes + 14, uniq_crashes + 19));
            } else if (line.contains("last uniq hang")) {
                int last_uniq_hang = line.indexOf(":");
                int uniq_hangs = line.indexOf("uniq hangs :");
                returnEntity.setLast_uniq_hang(line.substring(last_uniq_hang + 1, last_uniq_hang + 30));
                returnEntity.setUniq_hangs(line.substring(uniq_hangs + 12, uniq_hangs + 16));
            } else if (line.contains("now processing")) {
                int now_processing = line.indexOf(":");
                int map_density = line.indexOf("map density :");
                returnEntity.setNow_processing(line.substring(now_processing + 1, now_processing + 12));
                returnEntity.setMap_density(line.substring(map_density + 13, map_density + 25));
            } else if (line.contains("paths timed out")) {
                int paths_timed_out = line.indexOf(":");
                int count_coverage = line.indexOf("count coverage :");
                returnEntity.setPaths_timed_out(line.substring(paths_timed_out + 1, paths_timed_out + 12));
                returnEntity.setCount_coverage(line.substring(count_coverage + 16, count_coverage + 35));
            } else if (line.contains("now trying")) {
                int now_trying = line.indexOf(":");
                int favored_paths = line.indexOf("favored paths :");
                returnEntity.setNow_trying(line.substring(now_trying + 1, now_trying + 18));
                returnEntity.setFavored_paths(line.substring(favored_paths + 15, favored_paths + 28));
            } else if (line.contains("stage execs")) {
                int stage_execs = line.indexOf(":");
                int new_edges_on = line.indexOf("new edges on :");
                returnEntity.setStage_execs(line.substring(stage_execs + 1, stage_execs + 18));
                returnEntity.setNew_edges_on(line.substring(new_edges_on + 14, new_edges_on + 31));
            } else if (line.contains("total execs")) {
                int total_execs = line.indexOf(":");
                int total_crashes = line.indexOf("total crashes :");
                returnEntity.setTotal_execs(line.substring(total_execs + 1, total_execs + 5));
                returnEntity.setTotal_crashes(line.substring(total_crashes + 15, total_crashes + 31));
            } else if (line.contains("exec speed") && line.contains("total tmouts")) {
                int exec_speed = line.indexOf(":");
                int total_tmouts = line.indexOf("total tmouts :");
                returnEntity.setExec_speed(line.substring(exec_speed + 1, exec_speed + 25));
                returnEntity.setTotal_tmouts(line.substring(total_tmouts + 14, total_tmouts + 27));
            } else if (line.contains("bit flips")) {
                int bit_flips = line.indexOf(":");
                int levels = line.indexOf("levels :");
                returnEntity.setBit_flips(line.substring(bit_flips + 1, bit_flips + 18));
                returnEntity.setLevels(line.substring(levels + 8, levels + 12));
            } else if (line.contains("byte flips")) {
                int byte_flips = line.indexOf(":");
                int pending = line.indexOf("pending :");
                returnEntity.setByte_flips(line.substring(byte_flips + 1, byte_flips + 16));
                returnEntity.setPending(line.substring(pending + 9, pending + 12));
            } else if (line.contains("arithmetics")) {
                int arithmetics = line.indexOf(":");
                int pend_fav = line.indexOf("pend fav :");
                returnEntity.setArithmetics(line.substring(arithmetics + 1, arithmetics + 16));
                returnEntity.setPend_fav(line.substring(pend_fav + 10, pend_fav + 16));
            } else if (line.contains("known ints")) {
                int known_ints = line.indexOf(":");
                int own_finds = line.indexOf("own finds :");
                returnEntity.setKnown_ints(line.substring(known_ints + 1, known_ints + 16));
                returnEntity.setOwn_finds(line.substring(own_finds + 11, own_finds + 14));
            } else if (line.contains("dictionary") && line.contains("imported")) {
                int dictionary = line.indexOf(":");
                int imported = line.indexOf("imported :");
                returnEntity.setDictionary(line.substring(dictionary + 1, dictionary + 16));
                returnEntity.setImported(line.substring(imported + 10, imported + 16));
            } else if (line.contains("havoc")) {
                int havoc = line.indexOf(":");
                int stability = line.indexOf("stability :");
                returnEntity.setHavoc(line.substring(havoc + 1, havoc + 18));
                returnEntity.setStability(line.substring(stability + 11, stability + 20));
            } else if (line.contains("trim")) {
                send = true;
                int trim = line.indexOf(":");
                returnEntity.setTrim(line.substring(trim + 1, trim + 10));
            }
        } catch (Exception e) {
            log.error("Formarting params error!");
            throw new AflnetException("Formarting params error!");
        }
        return returnEntity;
    }

    public String parse(CmdStartParams cmdStartParams) throws AflnetException {
        System.out.println("cmd is :" + cmdStartParams);
        if (cmdStartParams == null) {
            throw new AflnetException("CmdStartParams is null !");
        }
        StringBuilder cmd = new StringBuilder();
        if (cmdStartParams.getNetinfo() != "" && cmdStartParams.getNetinfo() != null) {
            cmd.append(" -N " + cmdStartParams.getNetinfo() + " ");
        } else {
            log.error("There is no netinfo !");
            throw new AflnetException("There is no netinfo !");
        }
        if (cmdStartParams.getProtopcol() != "") {
            cmd.append(" -P " + cmdStartParams.getProtopcol() + " ");
        }
        if (cmdStartParams.getWaiting() != 0) {
            cmd.append(" -D " + cmdStartParams.getWaiting() + " ");
        }
        if (cmdStartParams.getNsname() != "") {
            cmd.append(" -e " + cmdStartParams.getNsname() + " ");
        }
        if (cmdStartParams.getKill()) {
            cmd.append(" -K ");
        }
        if (cmdStartParams.getAwareMode()) {
            cmd.append(" -E " + " ");
        }
        if (cmdStartParams.getRegionMutationOperators()) {
            cmd.append(" -R " + " ");
        }
        if (cmdStartParams.getFalseNegativeMode()) {
            cmd.append(" -F " + " ");
        }
        if (cmdStartParams.getScript() != "") {
            cmd.append(" -c " + cmdStartParams.getScript() + " ");
        }
        if (cmdStartParams.getStateSelectionAlgo() != 0) {
            cmd.append(" -q " + cmdStartParams.getStateSelectionAlgo() + " ");
        }
        if (cmdStartParams.getSeedSelectionAlgo() != 0) {
            cmd.append(" -s " + cmdStartParams.getSeedSelectionAlgo() + " ");
        }
        return cmd.toString();
    }

    @Transactional(rollbackFor = MysqlException.class)
    public void dataBackUpTransaction(String caller, List<String> out, List<String> error, String missionName) {
        int missionId = SystemRunningParams.kittyMissionId;
        try {
            /*kitty结果存入数据库*/
            KittyResult kittyResult = new KittyResult(missionId, out.toString(), error.toString());
            GlobalClass.kittyMapper.kittyResultsBackup(kittyResult);
        } catch (Exception e) {
            e.printStackTrace();
            log.error("kitty backup error!");
        }
//        /*kitty的协议模板方法的日志和数据包文件存入数据库*/
//        if (caller.equals("protocolTemplate")) {
////            GlobalClass.kittyFuzzPersistenceService.KittyPackagesBackup(missionId);
//            GlobalClass.missionInfoMapper.insertMission(new MissionInfo(missionId, 2, new Date(), missionName,));
//        } else {
//            /*存入missionInfo信息*/
//            GlobalClass.missionInfoMapper.insertMission(new MissionInfo(missionId, 3, new Date(), missionName));
//        }
    }
}